//~~ dealer.cc ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
 *  description
 *
 * @author:    Andy Pippin   <pippin@cs.ucsb.edu>
 * @version:   $Date: 2004/06/03 14:19:39 $   $Revision: 1.17 $
 * @requires:   
 */


//==[ Identity ]================================================================

#define  __DEALER_CC__
static char* srcIdentity = 
    "$Id: dealer.cc,v 1.17 2004/06/03 14:19:39 pippin Exp $";


//==[ Includes ]================================================================

#include  <errno.h>
#include  <fcntl.h>
#include  "mpi.h"
#include  <stdarg.h>
#include  <stdio.h>
#include  <string.h>
#include  <sys/stat.h>
#include  <sys/time.h>
#include  <sys/types.h>
#include  <unistd.h>
#include  <list>

#include  "john-mpi.hh"
#include  "dealer.hh"
#include  "trace.hh"
#include  "workmsg.hh"
#include  "john-1.6/src/worker.h"


//==[ Typedefs/Structures/Enums ]===============================================

//==[ Constants ]===============================================================

//==[ Variables ]===============================================================

namespace  john_mpi
{
    // Yes, these are globals.  Icky, huh?
    //
    static int  hours   = 0;
    static int  minutes = 0;
    static int  seconds = 0;
    static int  cracks  = 0;

    static int  results = -1;
};


//==[ Classes ]=================================================================

//==[ Prototypes ]==============================================================

//------------------------------------------------------------------------------
/*
 *  Create a work message.  This method uses a mutex lock to prevent two
 *  threads from stepping all over each other.
 *
 * @param   workList  The list of work orders.
 * @param   filename  The name of the input (password) file.
 * @return  The number of work messages.
 */
	static int
CreateWorkList( 
    std::list<WorkMsg*>&  workList,
    char*  filename
);


//------------------------------------------------------------------------------
/*
 *  Process all the requests.
 *
 * @param   workList    The list of work orders.
 * @param   resultList  The list of results.
 * @param   output      Where to log the results.
 * @return  The exit code.
 */
	static int
Deal(
    std::list<WorkMsg*>&  workList,
    std::list<WorkMsg*>&  resultList
);


//------------------------------------------------------------------------------
/*
 *  Process all of the results.
 *
 * @param   resultList  The list of results.
 * @param   output      Where to log the results.
 */
	static void
ProcessResults(
    std::list<WorkMsg*>&  resultList
);


//------------------------------------------------------------------------------
/*
 *  Print information to the result log file.
 *
 * @param   output   Where to display the workmsg line.
 * @param   fmt      The string to print to the result file.
 * @param   ...      Additional arguments.
 */
	static void
PrintResult(
    char*  fmt,
    ...
);


//------------------------------------------------------------------------------
/*
 *  Display an individual result of the form <(user) "password" == result.>
 *
 * @param   workMsg  The list of results.
 * @param   output   Where to display the workmsg line.
 */
	static void
ShowResult(
    WorkMsg*  workMsg,
    int       output
);


//------------------------------------------------------------------------------
/*
 *  Add the elapsed time to the global dealer wait time.
 *
 * @param   start  Start time of the wait.
 * @param   stop   Stop time of the wait.
 */
	static void
AddDealerWaitTime(
    struct timeval*  start,
    struct timeval*  stop
);


//------------------------------------------------------------------------------
/*
 *  Wrapper around strcmp for the quick sort (qsort).
 *
 * @param   lhsString  The left-hand side string.
 * @param   rhsString  The right-hand side string.
 * @return  The result of the strcmp.
 */
	static int
qsort_strcmp(
    const void* lhsString,
    const void* rhsString
);


//==[ Implementation ]==========================================================

	void
Dealer::Initialize(
    void
){
    this->passwordFilename = john_mpi::DEFAULT_PasswordFilename;
    this->resultsFilename  = john_mpi::DEFAULT_ResultsFilename;
    this->resultsFilename  = john_mpi::DEFAULT_WordlistFilename;

    if( john_mpi::dealerWaitTime )
        memset( &john_mpi::totalDealerWaitTime, 0,
				    sizeof(john_mpi::totalDealerWaitTime) );

    return;
}


//------------------------------------------------------------------------------

Dealer::Dealer(
    void
){
    Initialize();
    return;
}


//------------------------------------------------------------------------------

Dealer::~Dealer(
    void
){
}


//------------------------------------------------------------------------------

	int
Dealer::start(
){
    return( start( this->passwordFilename,
		   this->resultsFilename,
		   this->wordlistFilename ) );
}


//------------------------------------------------------------------------------

	void
Dealer::stop(
){
    int const  numWorkers = MPI::COMM_WORLD.Get_size();
    for( register int worker=1; worker < numWorkers; ++worker )
        WorkMsg::SendShutdown( worker );

    return;
}


//~~ Getters ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

char*  Dealer::getPasswordFilename( void )
    { return( this->passwordFilename ); }

char*  Dealer::getResultsFilename( void )
    { return( this->resultsFilename ); }


//~~ Setters ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Dealer::setPasswordFilename( char* passwordFilename )
    { this->passwordFilename = passwordFilename;  return; }

void Dealer::setResultsFilename( char* resultsFilename )
    { this->resultsFilename = resultsFilename;  return; }


//------------------------------------------------------------------------------

	char*
Dealer::toString(
    char*  prefix,
    char*  label
){
    static std::string  result;
    static char  line[ 256 ];

    // Initialize and fix NULLs.
    //
    result.erase();
    if( NULL == prefix )  prefix = "";
    if( NULL == label  )  label  = "Dealer";

    snprintf( line, sizeof(line), "%s<%s>\n", prefix, label );
    result += line;


    snprintf( line, sizeof(line), "%s</%s>\n", prefix, label );
    result += line;

    return( (char*)result.c_str() );
}


//------------------------------------------------------------------------------

	int
Dealer::start(
    char*  passwordFilename,
    char*  resultsFilename,
    char*  wordlistFilename
){
    // Variables.
    //
    int       exitCode = 0;			// Result of this method.
    int       numMsgs  = 0;			// Number of messages.
    FILE*     input    = NULL;			// Input file (passwords).
    int       results  = -1;			// Output file (results).
    WorkMsg*  workMsg  = NULL;			// A message of joy.

    time_t  rightNow;				// Right.... NOW!
    char  timeBuffer[ 81 ];			// Printable time.
    struct tm  startClock;			// Begin of wall clock time.
    struct tm  stopClock;			// Finish of wall clock time.

    static std::list<WorkMsg*>  workList;	// List of work messages.
    static std::list<WorkMsg*>  resultList;	// List of results.


    // Check parameters.
    //
    if( NULL == passwordFilename )
	this->passwordFilename = john_mpi::DEFAULT_PasswordFilename;
    if( NULL == resultsFilename )
	this->passwordFilename = john_mpi::DEFAULT_ResultsFilename;
    if( NULL == wordlistFilename )
	this->wordlistFilename = john_mpi::DEFAULT_WordlistFilename;


    // Initialization
    //
    rightNow = time(0);
    (void)localtime_r( &rightNow, &startClock );


    // Create the work list and record the size.
    //
    numMsgs = CreateWorkList( workList, passwordFilename );
    trace::clear();
    trace::log( __FILE__, __LINE__, "numMsgs = %d\n", numMsgs );


    // Set the results file.
    //
    int    flags = O_CREAT | O_TRUNC | O_SYNC | O_WRONLY;
    mode_t mode  = S_IRUSR | S_IWUSR;

    // Standard out?
    //
    if( 0 == strcmp( "-", resultsFilename ) )
    {
	john_mpi::results = 1;
    }
    else if( -1 == (john_mpi::results = open( resultsFilename, flags, mode )) )
    {
	if( 0 < john_mpi::verboseLevel )
	    fflush( stdout );
	    fprintf( stderr, "ERROR: Could not open \"%s\": %s\n",
					resultsFilename, strerror( errno ) );
	    fflush( stderr );

	fclose( input );

			//    *************************
	return( 2 );    //   ***  DOES NOT CONTINUE  ***
			//    *************************
    }
    else if( 0 < john_mpi::verboseLevel )
    {
	printf( "Results are being logged to: \"%s\"\n", resultsFilename );
    }


    // If the verbosity is high enough, display the messages.
    //
    if( 3 < john_mpi::verboseLevel )
    {
	PrintResult( "<WorkMsgList size=\"%d\">\n", workList.size() );

	std::list<WorkMsg*>::iterator  currMsg = workList.begin();
	std::list<WorkMsg*>::iterator  lastMsg = workList.end();

	for( ; currMsg != lastMsg; ++currMsg )
	    PrintResult( "%s", (*currMsg)->toString( "  " ) );

	PrintResult( "</WorkMsgList>\n" );
    }


    // Announce what is going on.
    //
    if( 0 < john_mpi::verboseLevel )
    {
	printf( "There are %d work orders and %d workers.\n",
	    workList.size(), MPI::COMM_WORLD.Get_size() -1  );

	if( 1 < john_mpi::verboseLevel )
	{
	    printf( "    Password file: %s\n", passwordFilename );
	    printf( "     Results file: %s\n", resultsFilename );
	    printf( "    Wordlist file: %s\n", wordlistFilename );
	}
    }

    // Send the starting statistics to the results file.
    //
    strftime( timeBuffer, sizeof(timeBuffer),
	"====[ Start: %a, %b %d %Y @ %H:%M:%S ]========================================================================", &startClock );
    timeBuffer[ 80 ] = '\0';

    PrintResult( "%s\n", timeBuffer );

    PrintResult( "\nConfiguration:\n" );
    PrintResult( "      Work orders: %d\n", workList.size() );
    PrintResult( "          Workers: %d\n", MPI::COMM_WORLD.Get_size()-1 );
    PrintResult( "    Password file: %s\n", passwordFilename );
    PrintResult( "     Results file: %s\n", resultsFilename );
    PrintResult( "    Wordlist file: %s\n", wordlistFilename );
    PrintResult( "\n\n-- Results ---------------------------------------------------------------------\n\n" );


    // Do the actual dealing.
    //
    exitCode = Deal( workList, resultList );


    // Process the results.
    //
    ProcessResults( resultList );
    PrintResult( "\n--------------------------------------------------------------------------------\n\n" );


    // Compute wall clock time.
    //
    rightNow = time(0);
    (void)localtime_r( &rightNow, &stopClock );

    // Compute the time.
    //
    int  wallHours   = stopClock.tm_hour - startClock.tm_hour;
    int  wallMinutes = stopClock.tm_min  - startClock.tm_min;
    int  wallSeconds = stopClock.tm_sec  - startClock.tm_sec;
    //
    if( 0 > wallSeconds )
    {
	wallSeconds += 60;
	wallMinutes -= 1;
    }
    if( 0 > wallMinutes )
    {
	wallMinutes += 60;
	wallHours   -= 1;
    }


    // -- Compute final statistics.
    //
    int totalWallclockSeconds = 3600 * wallHours
			      + 60   * wallMinutes
			      +        wallSeconds;

    float  cpsf;
    if( 0 != totalWallclockSeconds )
	cpsf = (float)john_mpi::cracks / (float)totalWallclockSeconds;
    else
        cpsf = 0.0f;

    int cps = (int)cpsf;


    // -- Display final statistics.
    //
    PrintResult( "\nFinal results:\n" );
    PrintResult( "       Wall clock: %02d:%02d:%02d\n",
			wallHours, wallMinutes, wallSeconds );
    PrintResult( "     Compute time: %02d:%02d:%02d\n",
			john_mpi::hours, john_mpi::minutes, john_mpi::seconds );
    PrintResult( "    Cracks/second: %d\n", cps );


    // Display the amount of time the dealer was waiting?
    //
    if( john_mpi::dealerWaitTime )
    {
	PrintResult( "    Dealer waited: %d.%02d seconds\n\n",
		john_mpi::totalDealerWaitTime.tv_sec,
		(john_mpi::totalDealerWaitTime.tv_usec + 5000) / 10000 );
    }


    strftime( timeBuffer, sizeof(timeBuffer),
	"====[ Finish: %a, %b %d %Y @ %H:%M:%S ]========================================================================", &stopClock );
    timeBuffer[ 80 ] = '\0';

    PrintResult( "%s\n", timeBuffer );

    // Close results file.
    //
    if( (1 != john_mpi::results) && (-1 != john_mpi::results) )
	 close( john_mpi::results );


    // Announce timings.
    //
    if( 1 == john_mpi::verboseLevel )
    {
	printf( "Done in %02d:%02d:%02d (hh:mm:ss).  See results file (%s).\n",
	    wallHours, wallMinutes, wallSeconds, resultsFilename );
    }
    else if( 1 < john_mpi::verboseLevel )
    {
	printf( "Done in %02d:%02d:%02d (hh:mm:ss).\n",
	    wallHours, wallMinutes, wallSeconds );
    }

    if( 1 < john_mpi::verboseLevel )
    {
	printf( "Total compute time was: %02d:%02d:%02d\n",
			john_mpi::hours, john_mpi::minutes, john_mpi::seconds );
    }



    return( exitCode );
}


//******************************************************************************
//**************************** Local Static Methods ****************************
//******************************************************************************

	static int
CreateWorkList( 
    std::list<WorkMsg*>&  workList,
    char*  filename
){
    int       result = 0;
    FILE*     passwdFile = NULL;
    char*     passwdEntry[ 8*1024 ] = { NULL };
    int       numEntries = 0;
    char      readBuffer[ BUFSIZ ];
    char*     encrypted[ WorkMsg::MAX_PASSWORDS ];
    WorkMsg*  workMsg = NULL;


    // Open the password file.
    //
    if( NULL == (passwdFile = fopen( filename, "r" )) )
    {
	if( 0 < john_mpi::verboseLevel )
	{
	    fflush( stdout );
	    fprintf( stderr, "ERROR: Could not open \"%s\": %s\n",
					filename, strerror( errno ) );
	    fflush( stderr );
	}
	result = 0;
    }


    //  Read in the entries from the password file and store them in an array
    //  so they can be sorted.  Toss out "*" or "x" passwords, and strip off
    //  the "expired-" if it exists.
    //
    else
    {
	while( NULL != fgets( readBuffer, sizeof(readBuffer), passwdFile ) )
	{
	    // Find the password.
	    //
	    char  *pwd, *colon;
	    if( (NULL != (pwd = strchr( readBuffer, ':' )))
	     && (NULL != (colon = strchr( ++pwd, ':' ))) )
	    {
	        *colon = '\0';

		// Create a work message only if the password is something.
		//
		if( 8 < strlen( pwd ) )
		{
		    passwdEntry[ numEntries++ ] = strdup( pwd );
		}

	    }  // Found the password

	}
	fclose( passwdFile );	 // Close the file.


	// Sort the list.
	//
	qsort( passwdEntry, numEntries, sizeof(char*), qsort_strcmp );


        // Now create the work messages.
	//
	int  entry = 0;
	int  numInMessage;
	char  *base, *peek;

	while( entry < numEntries )	 // First
	{
	    base = passwdEntry[ entry ];
	    numInMessage = 1;
	    peek = passwdEntry[ entry + numInMessage ];

	    while( ((entry+numInMessage) < numEntries)
	        && (0 == strncmp( base, peek, 2 ) ) )
		peek = passwdEntry[ entry + (++numInMessage) ];

	    workMsg = new WorkMsg( numInMessage, &(passwdEntry[entry]) );
	    workList.insert( workList.end(), workMsg );

	    entry += numInMessage;

	} // End while

    } // End else.

    return( workList.size() );
}


//------------------------------------------------------------------------------

	static int
Deal(
    std::list<WorkMsg*>&  workList,
    std::list<WorkMsg*>&  resultList
){
    int        exitCode = 0;			// Result of method.
    WorkMsg*   recvMsg;				// Message received.
    int        msgSource;			// Message source (rank).

    struct timeval  startWait;			// Begin of dealer waiting.
    struct timeval  stopWait;			// Finish of dealer waiting.


    // Initialize.
    //
    std::list<WorkMsg*>::iterator  currMsg = workList.begin();
    std::list<WorkMsg*>::iterator  lastMsg = workList.end();

    if( 1 < john_mpi::verboseLevel )
	fputc( '\n', stdout );

    // Go through the list.
    //
    for( ; currMsg != lastMsg; ++currMsg )
    {
	// Get the incoming message.
	if( john_mpi::dealerWaitTime )
	    gettimeofday( &startWait, NULL );

	recvMsg = WorkMsg::Recv();

	if( john_mpi::dealerWaitTime )
	{
	    gettimeofday( &stopWait, NULL );
	    AddDealerWaitTime( &startWait, &stopWait );
	}


	// Handle message.
	//
	if( NULL == recvMsg )
	{
	    exitCode = -2;
	    break;
	}
	msgSource = recvMsg->getSource();
	trace::log( __FILE__, __LINE__,
		"Got response: %4d, from: %2d\n",
		recvMsg->getSequence(), msgSource );


	// Save the result.
	//
	int num = recvMsg->getNumEntries();
	if( 0 != num )
	{
	    resultList.insert( resultList.end(), recvMsg );
	    ShowResult( recvMsg, john_mpi::results );

	    if( (1 < john_mpi::verboseLevel)
	     && (1 != john_mpi::results) )
	    {
		ShowResult( recvMsg, 1 );
	    }
	}



	if( false == (*currMsg)->Send( msgSource ) )
	{
	    exitCode = -3;
	    break;
	}
	else
	{
	    trace::log( __FILE__, __LINE__,
		    "Sending item: %4d, to:   %2d\n",
		    recvMsg->getSequence(), msgSource );
	}

    }

    // Shut down the workers.
    //
    int  numWorkers = MPI::COMM_WORLD.Get_size();
    for( int worker=1; worker < numWorkers; ++worker )
    {
	if( NULL != (recvMsg = WorkMsg::Recv()) )
	{
	    // Save the result.
	    //
	    int num = recvMsg->getNumEntries();
	    if( 0 != num )
	    {
		resultList.insert( resultList.end(), recvMsg );
		ShowResult( recvMsg, john_mpi::results );

		if( (1 < john_mpi::verboseLevel)
		 && (1 != john_mpi::results) )
		{
		    ShowResult( recvMsg, 1 );
		}
	    }

	    WorkMsg::SendShutdown( recvMsg->getSource() );
	}
    }

    if( 1 < john_mpi::verboseLevel )
	fputc( '\n', stdout );

    return( exitCode );
}


//------------------------------------------------------------------------------

	static void
ProcessResults(
    std::list<WorkMsg*>&  resultList
){
    int*      computeTime = NULL;	// Compute time array.
    int       thisCrack;		// How many seconds on this crack.

    int       numPasswords;		// Number of passwords in this msg.
    char*     en;			// The encrypted string.
    char*     de;			// The decrypted value.


    // List iterators.
    //
    std::list<WorkMsg*>::iterator  workMsg = resultList.begin();
    std::list<WorkMsg*>::iterator  endList = resultList.end();


    for( ; workMsg != endList; ++workMsg )
    {
	// Save the compute time.
	//
	computeTime = (*workMsg)->getCompute();
	john_mpi::hours   += computeTime[ WorkMsg::TIME_Hours   ];
	john_mpi::minutes += computeTime[ WorkMsg::TIME_Minutes ];
	john_mpi::seconds += computeTime[ WorkMsg::TIME_Seconds ];

	thisCrack = 3600 * computeTime[ WorkMsg::TIME_Hours   ]
		  + 60   * computeTime[ WorkMsg::TIME_Minutes ]
		  +        computeTime[ WorkMsg::TIME_Seconds ];
	john_mpi::cracks  += ( (*workMsg)->getCracksPerSecond() * thisCrack );
    }

    return;
}


//------------------------------------------------------------------------------

	static void
ShowResult(
    WorkMsg*  workMsg,
    int       output
){
    char*  en;
    char*  de;
    char   writeBuffer[ 1024 ];

    if( NULL != workMsg )
    {
	for( int passwd=0; passwd < workMsg->getNumEntries(); ++passwd )
	{
	    en = workMsg->getEncrypted( passwd );
	    de = workMsg->getDecrypted( passwd );

	    // Did we break it?
	    //
	    if( (NULL != de) && ('\0' != *de) )
	    {
		snprintf( writeBuffer, 1024, "    \"%s\" == \"%s\"\n", en, de );
	    }

	    // Nope.
	    else
	    {
		snprintf( writeBuffer, 1024, "    \"%s\"  not cracked.\n", en );
	    }
	}

	write( output, writeBuffer, strlen( writeBuffer ) );
    }

    return;
}


//------------------------------------------------------------------------------

	static void
PrintResult(
    char*  fmt,
    ...
){
    char writeBuffer[ 1024 ];
    int  bufferLen;

    if( (-1 != john_mpi::results)
     && (NULL != fmt) )
    {
        va_list  varArgs;
	va_start( varArgs, fmt );
	vsnprintf( writeBuffer, sizeof(writeBuffer), fmt, varArgs );
	va_end( varArgs );

	bufferLen = strlen( writeBuffer );
	(void)write( john_mpi::results, writeBuffer, bufferLen );
    }

    return;
}


//------------------------------------------------------------------------------

	static void
AddDealerWaitTime(
    struct timeval*  start,
    struct timeval*  stop
){
    time_t       sec;
    suseconds_t  usec;

    if( (NULL != start) && (NULL != stop) )
    {
        sec  = stop->tv_sec  - start->tv_sec;
	usec = stop->tv_usec - start->tv_usec;

	if( 0 > usec )
	{
	    sec -= 1;
	    usec += 1000000;
	}


	// Add it to the total time.
	//
	john_mpi::totalDealerWaitTime.tv_sec  += sec;
	john_mpi::totalDealerWaitTime.tv_usec += usec;

	while( 1000000 < john_mpi::totalDealerWaitTime.tv_usec )
	{
	    john_mpi::totalDealerWaitTime.tv_sec  += 1;
	    john_mpi::totalDealerWaitTime.tv_usec -= 1000000;
	}
    }

    return;
}


//------------------------------------------------------------------------------

	static int
qsort_strcmp(
    const void* arg1,
    const void* arg2
){
    return( strcmp( *(char**)arg1, *(char**)arg2 ) );
}

//==============================================================================
